import unittest
import json

from aes_bridge import encrypt_cbc, decrypt_cbc, encrypt_gcm, decrypt_gcm, encrypt_legacy, decrypt_legacy

class TestAesBridge(unittest.TestCase):
    pass

def add_test(name, method):
    # print(f"Adding test: {name}")
    setattr(TestAesBridge, name, method)

def load_dynamic_tests():
    with open('tests/test_data.json', 'r') as f:
        test_data = json.load(f)

    def test_encrypt_cbc_not_empty(value):
        def test(self):
            encrypted = encrypt_cbc(value, value)
            self.assertTrue(encrypted, "Encryption result should not be empty")
        return test

    def test_encrypt_gcm_not_empty(value):
        def test(self):
            encrypted = encrypt_gcm(value, value)
            self.assertTrue(encrypted, "Encryption result should not be empty")
        return test

    def test_encrypt_legacy_not_empty(value):
        def test(self):
            encrypted = encrypt_legacy(value, value)
            self.assertTrue(encrypted, "Encryption result should not be empty")
        return test

    def test_encrypt_decrypt_cbc(value):
        def test(self):
            encrypted = encrypt_cbc(value, value)
            decrypted = decrypt_cbc(encrypted, value)
            self.assertEqual(value, decrypted, "CBC encryption/decryption failed")
        return test

    def test_encrypt_decrypt_gcm(value):
        def test(self):
            encrypted = encrypt_gcm(value, value)
            decrypted = decrypt_gcm(encrypted, value)
            self.assertEqual(value, decrypted, "GCM encryption/decryption failed")
        return test

    def test_encrypt_decrypt_legacy(value):
        def test(self):
            encrypted = encrypt_legacy(value, value)
            decrypted = decrypt_legacy(encrypted, value)
            self.assertEqual(value, decrypted, "CBC encryption/decryption failed")
        return test

    def test_decrypt_cbc(encrypted, passphrase, result):
        def test(self):
            decrypted = decrypt_cbc(encrypted, passphrase)
            self.assertEqual(result, decrypted, "CBC decryption failed")
        return test

    def test_decrypt_gcm(encrypted, passphrase, result):
        def test(self):
            decrypted = decrypt_gcm(encrypted, passphrase)
            self.assertEqual(result, decrypted, "GCM decryption failed")
        return test

    def test_decrypt_legacy(encrypted, passphrase, result):
        def test(self):
            decrypted = decrypt_legacy(encrypted, passphrase)
            self.assertEqual(result, decrypted, "Legacy decryption failed")
        return test

    test_key = 'plaintext'
    for idx, test_case in enumerate(test_data.get('testdata', {}).get('plaintext', [])):
        test_case = test_case.encode('utf-8')
        add_test(f'test_0_{test_key}_encrypt_cbc_not_empty_{idx}', test_encrypt_cbc_not_empty(test_case))
        add_test(f'test_0_{test_key}_encrypt_gcm_not_empty_{idx}', test_encrypt_gcm_not_empty(test_case))
        add_test(f'test_0_{test_key}_encrypt_legacy_not_empty_{idx}', test_encrypt_legacy_not_empty(test_case))

        add_test(f'test_1_{test_key}_encrypt_decrypt_cbc_{idx}', test_encrypt_decrypt_cbc(test_case))
        add_test(f'test_1_{test_key}_encrypt_decrypt_gcm_{idx}', test_encrypt_decrypt_gcm(test_case))
        add_test(f'test_1_{test_key}_encrypt_decrypt_legacy_{idx}', test_encrypt_decrypt_legacy(test_case))

    test_key = 'hex'
    for idx, test_case in enumerate(test_data.get('testdata', {}).get('hex', [])):
        test_text = bytes.fromhex(test_case)
        add_test(f'test_2_{test_key}_encrypt_cbc_not_empty_{idx}', test_encrypt_cbc_not_empty(test_text))
        add_test(f'test_2_{test_key}_encrypt_gcm_not_empty_{idx}', test_encrypt_gcm_not_empty(test_text))
        add_test(f'test_2_{test_key}_encrypt_legacy_not_empty_{idx}', test_encrypt_legacy_not_empty(test_text))

        add_test(f'test_3_{test_key}_encrypt_decrypt_cbc_{idx}', test_encrypt_decrypt_cbc(test_text))
        add_test(f'test_3_{test_key}_encrypt_decrypt_gcm_{idx}', test_encrypt_decrypt_gcm(test_text))
        add_test(f'test_3_{test_key}_encrypt_decrypt_legacy_{idx}', test_encrypt_decrypt_legacy(test_text))

    for idx, test_case in enumerate(test_data.get('decrypt', [])):
        test_key = test_case.get('id', f'case_{idx}')
        plaintext = test_case.get('plaintext')
        hex_data = test_case.get('hex')
        passphrase = test_case.get('passphrase')
        encrypted_cbc = test_case.get('encrypted-cbc')
        encrypted_gcm = test_case.get('encrypted-gcm')
        encrypted_legacy = test_case.get('encrypted-legacy')

        if passphrase is None or (plaintext is None and hex_data is None) or \
           (encrypted_cbc is None and encrypted_gcm is None and encrypted_legacy is None):
            continue

        if plaintext is not None:
            plaintext = plaintext.encode('utf-8')
        elif hex_data is not None:
            plaintext = bytes.fromhex(hex_data)

        if encrypted_cbc is not None:
            add_test(f'test_9_decrypt_cbc_{test_key}', test_decrypt_cbc(encrypted_cbc, passphrase, plaintext))
        if encrypted_gcm is not None:
            add_test(f'test_9_decrypt_gcm_{test_key}', test_decrypt_gcm(encrypted_gcm, passphrase, plaintext))
        if encrypted_legacy is not None:
            add_test(f'test_9_decrypt_legacy_{test_key}', test_decrypt_legacy(encrypted_legacy, passphrase, plaintext))


load_dynamic_tests()

if __name__ == '__main__':
    unittest.main()



# This file is part of AesBridge - modern cross-language AES encryption library
# Repository: https://github.com/mervick/aes-bridge
#
# Copyright Andrey Izman (c) 2018-2025 <izmanw@gmail.com>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

import base64
from hashlib import md5
from os import urandom
from cryptography.hazmat.primitives.ciphers import Cipher, algorithms, modes
from cryptography.hazmat.backends import default_backend
from common import to_bytes, generate_random

BLOCK_SIZE = 16
KEY_LEN = 32
IV_LEN = 16


def encrypt_legacy(data: str | bytes, passphrase: str | bytes) -> bytes:
    """
    Encrypts plaintext using AES-256-CBC with OpenSSL-compatible output (Salted__ + salt + ciphertext),
    then base64-encodes the result.

    @param data: Plaintext data to encrypt (string or bytes)
    @param passphrase: Passphrase for key derivation
    @return: Encrypted data, base64-encoded as bytes
    """
    salt = generate_random(8)
    key, iv = __derive_key_and_iv(passphrase, salt)
    padded = __pkcs7_padding(data)
    cipher = Cipher(algorithms.AES(key), modes.CBC(iv), backend=default_backend()).encryptor()
    ciphertext = cipher.update(padded) + cipher.finalize()
    return base64.b64encode(b'Salted__' + salt + ciphertext)

def decrypt_legacy(data: str | bytes, passphrase: str | bytes) -> bytes:
    """
    Decrypts base64-encoded AES-CBC data with OpenSSL-compatible format (Salted__ + salt + ciphertext).

    @param data: Encrypted data, base64-encoded string or bytes
    @param passphrase: Passphrase for key derivation
    @type data: str | bytes
    @type passphrase: str | bytes
    @return: Decrypted plaintext as raw bytes
    @rtype: bytes
    """
    ct = base64.b64decode(data)
    if ct[:8] != b'Salted__':
        return b''
    salt = ct[8:16]
    key, iv = __derive_key_and_iv(passphrase, salt)
    cipher = Cipher(algorithms.AES(key), modes.CBC(iv), backend=default_backend()).decryptor()
    decrypted = cipher.update(ct[16:]) + cipher.finalize()
    return __pkcs7_trimming(decrypted)

def __pkcs7_padding(s: str | bytes) -> bytes:
    s = to_bytes(s)
    pad_len = BLOCK_SIZE - len(s) % BLOCK_SIZE
    return s + bytes([pad_len] * pad_len)

def __pkcs7_trimming(s) -> bytes:
    return s[:-s[-1]]

def __derive_key_and_iv(password: str | bytes, salt: bytes) -> tuple:
    d = d_i = b''
    password = to_bytes(password)
    while len(d) < KEY_LEN + IV_LEN:
        d_i = md5(d_i + password + salt).digest()
        d += d_i
    return d[:KEY_LEN], d[KEY_LEN:KEY_LEN + IV_LEN]



# This file is part of AesBridge - modern cross-language AES encryption library
# Repository: https://github.com/mervick/aes-bridge
#
# Copyright Andrey Izman (c) 2018-2025 <izmanw@gmail.com>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

from os import urandom
import hashlib


_nonce = int.from_bytes(urandom(8), 'big')

def generate_random(size: int) -> bytes:
    """ Generates random bytes of specified size. """
    global _nonce
    _nonce += 1
    nonce_bytes = _nonce.to_bytes(8, 'big')
    data = urandom(13) + nonce_bytes + urandom(13)
    return hashlib.sha256(data).digest()[:size]


def to_bytes(s: str | bytes) -> bytes:
    """ Convert string to bytes. """
    return s.encode('utf-8') if isinstance(s, str) else s



# This file is part of AesBridge - modern cross-language AES encryption library
# Repository: https://github.com/mervick/aes-bridge
#
# Copyright Andrey Izman (c) 2018-2025 <izmanw@gmail.com>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

import base64
from cryptography.hazmat.primitives import hashes
from cryptography.hazmat.primitives.kdf.pbkdf2 import PBKDF2HMAC
from cryptography.hazmat.primitives.ciphers import Cipher, algorithms, modes
from common import to_bytes, generate_random


def derive_key(passphrase: bytes, salt: bytes) -> bytes:
    return PBKDF2HMAC(
        algorithm=hashes.SHA256(),
        length=32,
        salt=salt,
        iterations=100_000,
    ).derive(passphrase)

def encrypt_gcm_bin(data: bytes | str, passphrase: bytes | str) -> bytes:
    """
    Encrypt data with AES-GCM.

    @param data: Data to encrypt
    @param passphrase: Encryption passphrase

    @return: Encrypted data in format:
             salt(16) + nonce(12) + ciphertext + tag(16)
    """
    passphrase = to_bytes(passphrase)
    plaintext = to_bytes(data)
    # salt = urandom(16)
    salt = generate_random(16)
    # nonce = urandom(12)
    nonce = generate_random(12)
    key = derive_key(passphrase, salt)

    cipher = Cipher(algorithms.AES(key), modes.GCM(nonce))
    encryptor = cipher.encryptor()
    ciphertext = encryptor.update(plaintext) + encryptor.finalize()

    return salt + nonce + ciphertext + encryptor.tag  # type: ignore

def decrypt_gcm_bin(data: bytes | str, passphrase: bytes | str) -> bytes:
    """
    Decrypt data encrypted by encrypt_gcm_bin().

    @param data: Encrypted data
    @param passphrase: Encryption passphrase
    """
    data = to_bytes(data)
    passphrase = to_bytes(passphrase)

    salt = data[:16]
    nonce = data[16:28]
    tag = data[-16:]
    ciphertext = data[28:-16]

    key = derive_key(passphrase, salt)

    cipher = Cipher(algorithms.AES(key), modes.GCM(nonce, tag))
    decryptor = cipher.decryptor()
    return decryptor.update(ciphertext) + decryptor.finalize()

def encrypt_gcm(data: bytes | str, passphrase: bytes | str) -> bytes:
    """
    Encrypt data using AES-GCM and encode result to base64.

    @param data: Data to encrypt
    @param passphrase: Encryption passphrase
    """
    return base64.b64encode(encrypt_gcm_bin(data, passphrase))

def decrypt_gcm(data: bytes | str, passphrase: bytes | str) -> bytes:
    """
    Decrypt base64 encoded data encrypted by encrypt_gcm().

    @param data: Base64 encoded encrypted data
    @param passphrase: Encryption passphrase
    """
    return decrypt_gcm_bin(base64.b64decode(to_bytes(data)), passphrase)



# This file is part of AesBridge - modern cross-language AES encryption library
# Repository: https://github.com/mervick/aes-bridge
#
# Copyright Andrey Izman (c) 2018-2025 <izmanw@gmail.com>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

import base64
from cryptography.hazmat.primitives import hashes, hmac, padding
from cryptography.hazmat.primitives.kdf.pbkdf2 import PBKDF2HMAC
from cryptography.hazmat.primitives.ciphers import Cipher, algorithms, modes
from common import to_bytes, generate_random


def derive_keys(passphrase: bytes, salt: bytes) -> tuple:
    kdf = PBKDF2HMAC(
        algorithm=hashes.SHA256(),
        length=64,  # 32 for AES key, 32 for HMAC key
        salt=salt,
        iterations=100_000,
    )
    key_material = kdf.derive(passphrase)
    return key_material[:32], key_material[32:]  # AES key, HMAC key

def encrypt_cbc_bin(data: bytes | str, passphrase: bytes | str) -> bytes:
    """
    Encrypts data using AES-CBC mode with HMAC authentication.

    @param data: Data to encrypt
    @param passphrase: Encryption passphrase

    @return: Encrypted data in format: salt (16 bytes) + IV (16 bytes) +
             ciphertext (variable length) + HMAC tag (32 bytes).
    """
    passphrase = to_bytes(passphrase)
    plaintext = to_bytes(data)
    salt = generate_random(16)
    iv = generate_random(16)
    aes_key, hmac_key = derive_keys(passphrase, salt)

    padder = padding.PKCS7(128).padder()
    padded = padder.update(plaintext) + padder.finalize()

    cipher = Cipher(algorithms.AES(aes_key), modes.CBC(iv))
    encryptor = cipher.encryptor()
    ciphertext = encryptor.update(padded) + encryptor.finalize()

    mac = hmac.HMAC(hmac_key, hashes.SHA256())
    mac.update(iv + ciphertext)
    tag = mac.finalize()

    return salt + iv + ciphertext + tag

def decrypt_cbc_bin(data: bytes | str, passphrase: bytes | str) -> bytes:
    """
    Decrypts data encrypted with encrypt_cbc_bin() function.

    @param data: Encrypted data in format from encrypt_cbc_bin():
                 salt (16) + IV (16) + ciphertext (N) + HMAC (32)
    @param passphrase: passphrase used for encryption
    """
    passphrase = to_bytes(passphrase)
    data = to_bytes(data)
    salt = data[:16]
    iv = data[16:32]
    tag = data[-32:]
    ciphertext = data[32:-32]

    aes_key, hmac_key = derive_keys(passphrase, salt)

    mac = hmac.HMAC(hmac_key, hashes.SHA256())
    mac.update(iv + ciphertext)
    mac.verify(tag)  # raises exception if tag invalid

    cipher = Cipher(algorithms.AES(aes_key), modes.CBC(iv))
    decryptor = cipher.decryptor()
    padded = decryptor.update(ciphertext) + decryptor.finalize()

    unpadder = padding.PKCS7(128).unpadder()
    return unpadder.update(padded) + unpadder.finalize()

def encrypt_cbc(data: bytes | str, passphrase: bytes | str) -> bytes:
    """
    Encrypts data and returns result as base64 encoded bytes.

    @param data: Data to encrypt
    @param passphrase: Encryption passphrase
    """
    return base64.b64encode(encrypt_cbc_bin(data, passphrase))

def decrypt_cbc(data: bytes | str, passphrase: bytes | str) -> bytes:
    """
    Decrypts base64 encoded data encrypted with encrypt_cbc().

    @param data: Base64 encoded encrypted data
    @param passphrase: Encryption passphrase
    """
    return decrypt_cbc_bin(base64.b64decode(data), passphrase)



# This file is part of AesBridge - modern cross-language AES encryption library
# Repository: https://github.com/mervick/aes-bridge
#
# Copyright Andrey Izman (c) 2018-2025 <izmanw@gmail.com>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

import sys
import os
sys.path.insert(0, os.path.join(os.path.dirname(__file__)))

from cbc import encrypt_cbc, decrypt_cbc, encrypt_cbc_bin, decrypt_cbc_bin
from gcm import encrypt_gcm, decrypt_gcm, encrypt_gcm_bin, decrypt_gcm_bin
from legacy import encrypt_legacy, decrypt_legacy

encrypt = encrypt_gcm
decrypt = decrypt_gcm

__all__       = ['encrypt',         'decrypt', 
                 'encrypt_cbc',     'decrypt_cbc', 
                 'encrypt_cbc_bin', 'decrypt_cbc_bin',
                 'encrypt_gcm',     'decrypt_gcm', 
                 'encrypt_gcm_bin', 'decrypt_gcm_bin',
                 'encrypt_legacy',  'decrypt_legacy']
__revision__  = "$Id$"
__author__    = "Andrey Izman"
__email__     = "izmanw@gmail.com"
__copyright__ = "Copyright 2018-2025 Andrey Izman"
__license__   = "MIT"

